import pytest
import os
import tempfile
import rmm.core as core
import pathlib as pl

test_mod_about_xml = """<?xml version="1.0" encoding="utf-8"?>
<ModMetaData>
    <name>Test Mod</name>
    <author>Spoons</author>
    <url>test.com</url>
    <supportedVersions>
        <li>1.0</li>
        <li>1.1</li>
        <li>1.2</li>
        <li>1.3</li>
    </supportedVersions>
<description>
</description>
<packageId>spoons.test</packageId>
</ModMetaData>
"""

test_mod_file_id = "115251734"


@pytest.fixture
def test_create_mod_dir():
    with tempfile.TemporaryDirectory() as path:
        prev_path = os.getcwd()
        os.chdir(path)

        folders = ["1.1", "1.2", "1.3", "About", "Assemblies", "Defs", "Languages"]
        for n in folders:
            os.mkdir(n)

        with open("About/About.xml", "w") as f:
            f.write(test_mod_about_xml)

        with open("About/PublishedFileId.txt", "w") as f:
            f.write(test_mod_file_id)

        os.chdir(prev_path)
        yield path


def test_create_mod(test_create_mod_dir):
    mod_dir = test_create_mod_dir
    test_mod = core.Mod.create_from_path(mod_dir)
    assert test_mod.name == "Test Mod"
    assert test_mod.author == "Spoons"
    assert test_mod.versions == ["1.0", "1.1", "1.2", "1.3"]
    assert test_mod.fp == mod_dir



#!/usr/bin/env python3

from __future__ import annotations

import xml.etree.ElementTree as ET
from multiprocessing import Pool
from pathlib import Path
from typing import Optional, cast
from dataclasses import dataclass
from collections import OrderedDict

import rmm.util as util
from rmm.exception import InvalidPackageHash
DEBUG = False


@dataclass
class Mod:
    def __init__(
        self,
        packageid: Optional[str] = None,
        before: Optional[list[str]] = None,
        after: Optional[list[str]] = None,
        incompatible: Optional[list[str]] = None,
        dirname: Optional[Path] = None,
        author: Optional[str] = None,
        name: Optional[str] = None,
        versions: Optional[list[str]] = None,
        steamid: Optional[int] = None,
        ignored: bool = False,
        repo_url: Optional[str] = None,
        workshop_managed: Optional[bool] = None,
        enabled: Optional[bool] = None,
    ):
        if packageid and isinstance(packageid, str):
            self.packageid = packageid.lower()
        else:
            self.packageid = packageid
        self.before = Mod.lowercase_set(before)
        self.after = Mod.lowercase_set(after)
        self.incompatible = incompatible
        self.dirname = dirname
        self.author = author
        self.name = name
        self.ignored = ignored
        self.steamid = steamid
        self.versions = versions
        self.repo_url = repo_url
        self.workshop_managed = workshop_managed
        self.enabled = enabled

    def title(self):
        return self.packageid if self.packageid else f"{self.name} by {self.author}"

    @staticmethod
    def lowercase_set(data):
        retval = set()
        try:
            for n in data:
                try:
                    retval.add(n.lower())
                except AttributeError:
                    continue
        except TypeError:
            return None
        return retval

    @staticmethod
    def create_from_workshorp_result(wr: WorkshopResult):
        return Mod(steamid=wr.steamid, name=wr.name, author=wr.author)

    @staticmethod
    def create_from_path(path) -> Optional[Mod]:
        try:
            tree = ET.parse(path / "About/About.xml")
            root = tree.getroot()
            try:
                packageid = cast(str, cast(ET.Element, root.find("packageId")).text)
            except AttributeError as e:
                name = util.element_grab("name", root)
                author = util.element_grab("author", root)
                if name and author:
                    packageid = "{}.{}".format(name.lower(), author.lower())
                else:
                    if DEBUG:
                        raise e
                    return None

            def read_steamid(path: Path) -> Optional[int]:
                try:
                    return int(
                        (path / "About" / "PublishedFileId.txt")
                        .read_text()
                        .strip()
                        .encode("ascii", errors="ignore")
                        .decode()
                    )
                except (OSError, ValueError, IOError) as e:
                    if DEBUG:
                        print(e)
                    return None

            def read_ignored(path: Path):
                try:
                    return (path / ".rmm_ignore").is_file()
                except (OSError) as e:
                    print(e)
                    return False

            return Mod(
                packageid=packageid,
                before=util.list_grab("loadAfter", root),
                after=util.list_grab("loadBefore", root),
                incompatible=util.list_grab("incompatibleWith", root),
                dirname=path.name,
                author=util.element_grab("author", root) if util.element_grab("author", root) else ', '.join(util.list_grab("authors", root)),
                name=util.element_grab("name", root),
                versions=util.list_grab("supportedVersions", root),
                steamid=read_steamid(path),
                ignored=read_ignored(path),
            )

        except OSError as e:
            # if not "Place mods here" in path.name:
            print(f"Ignoring {path}")
            # raise e
        except ET.ParseError as e:
            print(f"Ignoring {path}.\n\t{path}/About/About.xml contains invalid XML.")

    def __eq__(self, other):
        if isinstance(other, Mod):
            if self.packageid and other.packageid:
                return self.packageid == other.packageid
            if self.steamid and other.steamid:
                return self.steamid == other.steamid
        if isinstance(other, str) and self.packageid:
            return self.packageid.lower() == other.lower()
        if isinstance(other, int) and self.steamid:
            return self.steamid == other
        return NotImplemented

    def __hash__(self):
        if not self.packageid:
            raise InvalidPackageHash()
        return hash(self.packageid)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<Mod: '{self.packageid}'>"

    @staticmethod
    def list_to_dict(mods: list[Mod]):
        hm = dict()
        for n in mods:
            if not n.packageid:
                raise Exception("No package id in hashmap")
            hm[n.packageid] = n

        return hm


class ModFolder:
    @staticmethod
    def read(path: Path) -> list[Mod]:
        with Pool(16) as p:
            mods = cast(
                list[Mod],
                list(
                    filter(
                        None,
                        p.map(
                            Mod.create_from_path,
                            path.iterdir(),
                        ),
                    )
                ),
            )

        return mods

    @staticmethod
    def read_dict(path: Path):
        return Mod.list_to_dict(ModFolder.read(path))

    @staticmethod
    def search(path: Path, search_term) -> list[Mod]:
        return [
            r
            for r in ModFolder.read(path)
            if (isinstance(r.name, str) and search_term.lower() in r.name.lower())
            or (isinstance(r.author, str) and search_term.lower() in r.author.lower())
            or search_term == r.steamid
        ]

    @staticmethod
    def search_dict(path: Path, search_term) -> dict[str, Mod]:
        return Mod.list_to_dict(ModFolder.search(path, search_term))


EXPANSION_PACKAGES = [
    Mod(packageid="ludeon.rimworld", author="Ludeon", name="RimWorld"),
    Mod(packageid="ludeon.rimworld.ideology", author="Ludeon", name="Ideology"),
    Mod(packageid="ludeon.rimworld.royalty", author="Ludeon", name="Royalty"),
    Mod(packageid="ludeon.rimworld.biotech", author="Ludeon", name="Biotech"),
]



#!/usr/bin/env python3

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import cast, List
from xml.etree import ElementTree as ET
from rmm.steam import SteamDownloader

import rmm.util as util
from rmm.mod import EXPANSION_PACKAGES, Mod


class ModsConfig:
    def __init__(self, p: Path):
        if isinstance(p, str):
            p = Path(p)
        self.path = p.expanduser()
        try:
            self.element_tree = ET.parse(self.path)
        except OSError:
            print("Unable to read ModsConfig file at " + str(self.path))
            raise

        self.root = self.element_tree.getroot()
        try:
            enabled = cast(
                List[str],
                util.list_grab("activeMods", self.root),
            )
            # self.mods = [Mod(packageid=pid) for pid in enabled]
            self.mods = { pid:None for pid in enabled}
        except TypeError:
            print("Unable to parse activeMods in ModsConfig")
            raise
        self.version = util.element_grab("version", self.root)
        self.length = len(self.mods)
        try:
            self.expansions = util.list_grab("knownExpansions", self.root)
            if not self.expansions:
                self.expansions = []
        except TypeError:
            self.expansions = []
            pass

    def write(self):
        active_mods = self.root.find("activeMods")
        if not active_mods:
            return
        try:
            for item in list(active_mods.findall("li")):
                if active_mods:
                    active_mods.remove(item)
        except AttributeError:
            pass

        try:
            for mod in self.mods:
                new_element = ET.SubElement(active_mods, "li")
                new_element.text = mod
        except AttributeError:
            raise Exception("Unable to find 'activeMods' in ModsConfig")

        buffer = util.et_pretty_xml(self.root)

        try:
            with self.path.open("w+") as f:
                f.seek(0)
                f.write(buffer)
        except OSError:
            print("Unable to write ModsConfig")
            raise

    def enable_mod(self, m: Mod):
        self.mods[m.packageid] = None

    def disable_mod(self, m: Mod):
        if m.packageid in self.mods:
            del self.mods[m.packageid]

    def autosort(self, mods, config):
        import json

        import networkx as nx

        DG = nx.DiGraph()

        before_core = ["brrainz.harmony", "me.samboycoding.betterloading"]

        core = ["ludeon.rimworld"]

        expansion_load_order = [
            "ludeon.rimworld.royalty",
            "ludeon.rimworld.ideology",
        ]

        combined_load_order = before_core + core + expansion_load_order
        for n, pid in enumerate(combined_load_order):
            if not pid in self.mods:
                del combined_load_order[n]

        for k in range(0, len(combined_load_order)):
            for j in range(k + 1, len(combined_load_order)):
                DG.add_edge(combined_load_order[j], combined_load_order[k])

        populated_mods = {m.packageid: m for m in mods if m in self.mods}

        rules_path = Path(config.mod_path / "rupal.rimpymodmanagerdatabase/db/communityRules.json")
        if not rules_path.is_file():
            print("Downloading rules file\n")
            # import rmm.steam
            # rules_cache_path = Path(SteamDownloader.find_path()[1]/ "1847679158")
            # rmm.steam.SteamDownloader.download([1847679158])
            import rmm.manager
            rmm.manager.Manager(config).sync_mods([Mod( steamid=1847679158 )])


        with (
            config.mod_path / "rupal.rimpymodmanagerdatabase/db/communityRules.json"
        ).open("r", encoding="utf-8") as f:
            community_db = json.load(f)

        for pid, m in populated_mods.items():
            try:
                for j in community_db["rules"][m.packageid]["loadAfter"]:
                    if j:
                        try:
                            m.before.add(j)
                        except AttributeError:
                            m.before = set(j)
            except KeyError:
                pass
            try:
                for j in community_db["rules"][m.packageid]["loadBefore"]:
                    if j:
                        try:
                            m.after.add(j)
                        except AttributeError:
                            m.after = set(j)
            except KeyError:
                pass

        rocketman = False
        if "krkr.rocketman" in populated_mods:
            rocketman = True

        if (
            "murmur.walllight" in populated_mods
            and "juanlopez2008.lightsout" in populated_mods
        ):
            DG.add_edge("juanlopez2008.lightsout", "murmur.walllight")

        mods_for_removal = {
            n
            for n in expansion_load_order + before_core
            if n not in EXPANSION_PACKAGES + mods
            or n not in self.mods
        }

        for pid, m in populated_mods.items():
            if rocketman and pid != "krkr.rocketman":
                DG.add_edge("krkr.rocketman", pid)
            if not m in combined_load_order:
                for n in combined_load_order:
                    DG.add_edge(pid, n)
            if m.after:
                for a in m.after:
                    if a in self.mods:
                        DG.add_edge(a.lower(), pid)
            if m.before:
                for b in m.before:
                    if b in self.mods:
                        DG.add_edge(pid, b.lower())

        count = 0
        while True:
            try:
                sorted_mods = reversed(list(nx.topological_sort(DG)))
                self.mods = [n for n in sorted_mods]
                self.mods = util.list_loop_exclusion(self.mods, mods_for_removal)
                print("Auto-sort complete")

                print(
                    "Verifying state: {}".format(
                        "good" if self.verify_state(populated_mods) else "bad"
                    )
                )
                return
            except nx.exception.NetworkXUnfeasible:
                if count >= 10:
                    print("Unable to break cycles")
                    exit(0)
                print("Cycle found. Breaking load order cycle")
                cycle = nx.find_cycle(DG)
                print(cycle)
                DG.remove_edge(*cycle[0])
                count += 1

    def verify_state(self, mods: List[Mod]):
        if isinstance(mods, list):
            populated_mods = {m.packageid: m for m in mods if m.packageid in self.mods}
        elif isinstance(mods, dict):
            populated_mods = {
                m.packageid: m for m in mods.values() if m.packageid in self.mods
            }
        else:
            raise Exception("bad data type")

        incompatible = dict()
        for m in populated_mods.values():
            try:
                for n in m.incompatible:
                    incompatible[n] = None
            except TypeError:
                pass

        for n in populated_mods:
            if n in incompatible:
                return False

        return True



from contextlib import contextmanager
import re
import shutil
import subprocess
import sys
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Generator, Optional, cast, Union, List
from xml.dom import minidom


def platform() -> Optional[str]:
    return sys.platform


def execute(cmd) -> Generator[str, None, None]:
    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
        text=True,
        close_fds=True,
        shell=True,
    ) as proc:
        for line in iter(proc.stdout.readline, b""):
            yield line
            if (r := proc.poll()) is not None:
                if r != 0:
                    raise subprocess.CalledProcessError(r, cmd)
                break


def run_sh(cmd: str) -> str:
    # Will raise a CalledProcessError if non-zero return code
    return subprocess.check_output(cmd, text=True, shell=True).strip()


def copy(source: Path, destination: Path, recursive: bool = False):
    if recursive:
        shutil.copytree(source, destination)
    else:
        shutil.copy2(source, destination, follow_symlinks=True)


def move(source: Path, destination: Path):
    shutil.move(source, destination)


def remove(dest: Path):
    shutil.rmtree(dest)


def list_set_intersection(a: list, b: list) -> list:
    return list(set(a) & set(b))


def list_loop_intersection(a: list, b: list) -> list:
    return [value for value in a if value in b]


def list_loop_exclusion(a: list, b: list) -> list:
    return [value for value in a if value not in b]


def list_grab(element: str, root: ET.Element) -> Optional[List[str]]:
    try:
        return cast(
            Optional[List[str]],
            [n.text for n in cast(ET.Element, root.find(element)).findall("li")],
        )
    except AttributeError:
        return None


def element_grab(element: str, root: ET.Element) -> Optional[str]:
    try:
        return cast(ET.Element, root.find(element)).text
    except AttributeError:
        return None


def et_pretty_xml(root: ET.Element) -> str:
    return minidom.parseString(
        re.sub(
            r"[\n\t\s]*",
            "",
            (ET.tostring(cast(ET.Element, root), "utf-8").decode()),
        )
    ).toprettyxml(indent="  ", newl="\n")

def sanitize_path(path: Union[str, Path]):
    if isinstance(path, Path):
        path = str(path)

    if platform() == "win32":
        path.replace('"', "")

    return Path(path).expanduser()



# import networkx as nx
# import pyplot as plt


# class GraphAnalyzer:
#     @staticmethod
#     def graph(mods):

#         DG = nx.DiGraph()

#         ignore = ["brrainz.harmony", "UnlimitedHugs.HugsLib"]
#         for m in mods:
#             if m.after:
#                 for a in m.after:
#                     if a in mods:
#                         if not a in ignore and not m.packageid in ignore:
#                             DG.add_edge(a, m.packageid)
#             if m.before:
#                 for b in m.before:
#                     if b in mods:
#                         if not b in ignore and not m.packageid in ignore:
#                             DG.add_edge(m.packageid, b)

#         pos = nx.spring_layout(DG, seed=56327, k=0.8, iterations=15)
#         nx.draw(
#             DG,
#             pos,
#             node_size=100,
#             alpha=0.8,
#             edge_color="r",
#             font_size=8,
#             with_labels=True,
#         )
#         ax = plt.gca()
#         ax.margins(0.08)

#         print("topological sort:")
#         sorted = list(nx.topological_sort(DG))
#         for n in sorted:
#             print(n)

#         plt.show()
